<?php

namespace Trello\Models;

use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Log;
use Trello\API;
use Trello\Models\CustomFields\ListList;

final class Board extends Model
{
	protected $keyType = 'string';
	protected const LOGKEY = 'MTB';

	public Collection $customfields;
	private API $api;
	public ?BoardList $newopslist = NULL;
	public const newopslist = 'New Ops';

	/**
	 * Use the API to store/retrieve values
	 *
	 * @return API
	 * @todo This should be kept private, and we should call functions to retrieve the items we are interested in
	 */
	public function api(): API
	{
		return $this->api;
	}

	/**
	 * Retrieve a specific custom field
	 *
	 * @param string $name
	 * @return CustomField
	 * @throws \Exception
	 */
	public function customField(string $name): ?CustomField
	{
		return $this->customFields()->filter(function($item) use ($name) { return $item->name == $name;})->pop();
	}

	/**
	 * Get this boards custom fields.
	 *
	 * @return Collection
	 * @throws \Exception
	 */
	public function customFields(): Collection
	{
		if (! $this->customfields->count()) {
			$cf = $this->api->getCustomFields($this);

			// Check all our customFields are of the types we need:
			foreach ($cf as $key => $field) {
				if ($x=config('trello.customfields.'.$field->name)) {

					if ($field instanceof $x) {
						dump(['m'=>__METHOD__,'EXISTS','x'=>$x,'field'=>$field,'listlsit'=>$field instanceof ListList]);

						// If it is a list item, we need to check that the list has all the options
						if ($field instanceof ListList) {
							if (! $field->hasAllOptions()) {
								$this->api->createCustomFieldOptions($field,collect($field->missingOptions())->transform(function($item) { return ['pos'=>'bottom','value'=>['text'=>$item]]; }));
								//dd(['m'=>__METHOD__,'has failed?'=>collect($field->missingOptions())->transform(function($item,$key) { return ['pos'=>$key,'value'=>['text'=>$item]]; }),'x'=>$x]);
							}
						}

					} else {
						Log::error(sprintf('%s:Field [%s] is not of the right type [%s] (%s)',self::LOGKEY,$field->name,$field->type,$x));

						throw new \Exception(sprintf('Invalid field type: %s (%s), should be [%s]',$field->name,get_class($field),$x));
					}

					// Not one of our fields
				} else {
					Log::debug(sprintf('%s:Ignoring custom field [%s]',self::LOGKEY,$field->name));

					$cf->forget($key);
				}
			}

			// Any missing fields we need to create.
			foreach (collect(config('trello.customfields'))->filter(function($item,$key) use ($cf) { return $cf->pluck('attributes.name')->search($key) === FALSE; }) as $key => $field) {
				Log::debug(sprintf('%s:Creating field [%s] (%s)',self::LOGKEY,$key,$field));

				$o = new $field(new \stdClass);
				dump(['o'=>$o]);
				$cf->push($x=$this->api->createCustomField($this,$key,$o->trello_type));
				//dd(['M'=>__METHOD__,'addkey'=>$key,'addfield'=>$field,'x'=>$x]);

				// If it is a list, we'll need to create our options too
				if ($x instanceof ListList) {
					$this->api->createCustomFieldOptions($x,collect($x->missingOptions())->transform(function($item) { return ['pos'=>'bottom','value'=>['text'=>$item]]; }));
				}
			}

			$this->customfields = $cf;
		}

		return $this->customfields;
	}

	/**
	 * Retrieve the list that we use to create new cards in
	 *
	 * @return BoardList
	 */
	public function listNewOps(): BoardList
	{
		if (! $this->newopslist) {
			// Find our "New Opps" List in Trello
			$newoplist = $this->api
				->getLists($this)
				->filter(function($item) { return $item->name == self::newopslist; });

			// If list doesnt exist, we'll create it.
			if (! $newoplist->count()) {
				//$newoplist = $api->createList($b,$newopslist);
				dd([__METHOD__=>$newoplist,'todo'=>'create this']);

			} else {
				$this->newoplist = $newoplist->pop();
			}
		}

		return $this->newoplist;
	}

	/**
	 * Record the API that this board is stored/retrieved with
	 *
	 * @param API $api
	 * @return void
	 */
	public function setAPI(API $api): void
	{
		$this->api = $api;
		$this->customfields = collect();
	}
}